import { GLType } from "@thi.ng/api/typedarray";
import { illegalArgs } from "@thi.ng/errors/illegal-arguments";
import { unsupported } from "@thi.ng/errors/unsupported";
import { WasmDomModule } from "@thi.ng/wasm-api-dom";
import { ObjectIndex } from "@thi.ng/wasm-api/object-index";
import { compileModel } from "@thi.ng/webgl/buffer";
import { draw } from "@thi.ng/webgl/draw";
import { defShader } from "@thi.ng/webgl/shader";
import { defTexture } from "@thi.ng/webgl/texture";
import {
  $AttribUpdateSpec,
  $ModelSpec,
  $ShaderSpec,
  $TextureSpec,
  $UniformValue,
  $WebGLContextOpts,
  ImageType,
  ShaderAttribType,
  UniformType,
  WebGLPowerPreference
} from "./generated/api.js";
const WasmWebGLModule = {
  id: "webgl",
  deps: [WasmDomModule],
  factory: () => new WasmWebGL()
};
class WasmWebGL {
  parent;
  dom;
  contexts;
  models;
  shaders;
  textures;
  $ShaderSpec;
  $ModelSpec;
  $TextureSpec;
  $UniformValue;
  $AttribUpdateSpec;
  $WebGLContextOpts;
  async init(parent) {
    this.parent = parent;
    this.dom = this.parent.modules[WasmDomModule.id];
    const defIndex = (name) => new ObjectIndex({ name, logger: parent.logger });
    this.contexts = defIndex("ctx");
    this.shaders = defIndex("sha");
    this.models = defIndex("mod");
    this.textures = defIndex("tex");
    this.$AttribUpdateSpec = $AttribUpdateSpec(this.parent);
    this.$ModelSpec = $ModelSpec(this.parent);
    this.$ShaderSpec = $ShaderSpec(this.parent);
    this.$TextureSpec = $TextureSpec(this.parent);
    this.$UniformValue = $UniformValue(this.parent);
    this.$WebGLContextOpts = $WebGLContextOpts(this.parent);
    return true;
  }
  getImports() {
    return {
      canvasGLContext: (canvasID, optsAddr) => {
        const opts = this.$WebGLContextOpts.instance(optsAddr);
        const ctx = this.dom.elements.get(canvasID).getContext("webgl2", {
          alpha: !!opts.alpha,
          antialias: !!opts.antialias,
          depth: !!opts.depth,
          desynchronized: !!opts.desynchronized,
          failIfMajorPerformanceCaveat: !!opts.failIfMajorPerformanceCaveat,
          powerPreference: WebGLPowerPreference[opts.powerPreference].replace("_", "-"),
          premultipliedAlpha: !!opts.premultipliedAlpha,
          preserveDrawingBuffer: !!opts.preserveDrawingBuffer,
          stencil: !!opts.stencil
        });
        if (!ctx) unsupported("WebGL2 not supported");
        return this.contexts.addUnique(ctx);
      },
      createShader: (ctxID, addr) => {
        const buildAttribs = (spec2) => {
          const attribs = {};
          for (let att of spec2.attribs) {
            attribs[att.name.deref()] = ShaderAttribType[att.type];
          }
          return attribs;
        };
        const buildVarying = (spec2) => {
          const varyings = {};
          for (let v of spec2.varying) {
            varyings[v.name.deref()] = ShaderAttribType[v.type];
          }
          return varyings;
        };
        const buildUniforms = (spec2) => {
          const uniforms = {};
          for (let uni of spec2.uniforms) {
            uniforms[uni.name.deref()] = [
              UniformType[uni.type],
              uni.default[UniformType[uni.type]]
            ];
          }
          return uniforms;
        };
        const gl = this.contexts.get(ctxID);
        const spec = this.$ShaderSpec.instance(addr);
        const $spec = {
          vs: spec.vs.deref(),
          fs: spec.fs.deref(),
          attribs: buildAttribs(spec),
          varying: buildVarying(spec),
          uniforms: buildUniforms(spec)
        };
        const shader = defShader(gl, $spec, {
          logger: this.parent.logger
        });
        return this.shaders.add(shader);
      },
      createModel: (ctxID, addr) => {
        const buildAttribs = (specs) => {
          const attribs = {};
          for (let att of specs) {
            attribs[att.name.deref()] = {
              type: att.type,
              size: att.size,
              stride: att.stride,
              offset: att.offset,
              data: att.data[GLType[att.type].toLowerCase()]
            };
          }
          return attribs;
        };
        const buildUniforms = (specs) => {
          const uniforms = {};
          for (let uni of specs) {
            uniforms[uni.name.deref()] = uni.value[UniformType[uni.type]];
          }
          return uniforms;
        };
        const gl = this.contexts.get(ctxID);
        const spec = this.$ModelSpec.instance(addr);
        const modelSpec = {
          shader: this.shaders.get(spec.shader),
          attribs: buildAttribs(spec.attribs),
          uniforms: buildUniforms(spec.uniforms),
          mode: spec.mode,
          num: spec.num
        };
        if (spec.numInstances) {
          modelSpec.instances = {
            attribs: buildAttribs(spec.instances),
            num: spec.numInstances
          };
        }
        if (spec.textures.length) {
          modelSpec.textures = [...spec.textures].map(
            (id) => this.textures.get(id)
          );
        }
        return this.models.add(compileModel(gl, modelSpec));
      },
      createTexture: (ctxID, addr) => {
        const gl = this.contexts.get(ctxID);
        const spec = this.$TextureSpec.instance(addr);
        const img = spec.imgType !== ImageType.none ? spec.img[ImageType[spec.imgType]] : null;
        const tex = defTexture(gl, {
          width: spec.width,
          height: spec.height,
          depth: spec.depth,
          image: img,
          filter: spec.filter,
          format: spec.format,
          wrap: spec.wrap,
          target: spec.target,
          type: spec.type
        });
        return this.textures.add(tex);
      },
      uniformInt: this.uniformScalar.bind(this),
      uniformUInt: this.uniformScalar.bind(this),
      uniformFloat: this.uniformScalar.bind(this),
      uniformVec: (modelID, name, value, size) => this.uniformVec(modelID, name, "f32", value >> 2, size),
      uniformIVec: (modelID, name, value, size) => this.uniformVec(modelID, name, "i32", value >> 2, size),
      uniformUVec: (modelID, name, value, size) => this.uniformVec(modelID, name, "u32", value >> 2, size),
      updateAttrib: (modelID, name, addr) => {
        const model = this.models.get(modelID);
        const id = this.parent.getString(name);
        const attrib = model.attribs[id];
        if (!attrib) illegalArgs(`unknown attrib: ${id}`);
        const spec = this.$AttribUpdateSpec.instance(addr);
        attrib.buffer.setChunk(
          spec.data[GLType[spec.type].toLowerCase()],
          spec.offset
        );
      },
      draw: (modelID) => {
        draw(this.models.get(modelID));
      },
      clear: (ctxID, r, g, b, a) => {
        const gl = this.contexts.get(ctxID);
        gl.clearColor(r, g, b, a);
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
      }
    };
  }
  /**
   * Shared handler for setting a scalar uniform in a ModelSpec
   *
   * @param modelID
   * @param name
   * @param value
   */
  uniformScalar(modelID, name, value) {
    const spec = this.models.get(modelID);
    const id = this.parent.getString(name);
    spec.uniforms[id] = value;
  }
  uniformVec(modelID, name, type, addr, size) {
    const spec = this.models.get(modelID);
    const id = this.parent.getString(name);
    const val = this.parent[type].subarray(addr, addr + size).slice();
    spec.uniforms[id] = val;
  }
}
export {
  WasmWebGL,
  WasmWebGLModule
};
